import numpy as np
import math 

def cosmological_distances(z):
    
    # initialize constants
    H0 = 70
    WM = 0.315
    WV = 1.0 - WM - 0.4165/(H0*H0)  # Omega(vacuum) or lambda

    WR = 0.        # Omega(radiation)
    WK = 0.        # Omega curvaturve = 1-Omega(total)
    c = 299792.458 # velocity of light in km/sec
    Tyr = 977.8    # coefficent for converting 1/H into Gyr
    DTT = 0.5      # time from z to now in units of 1/H0
    DTT_Gyr = 0.0  # value of DTT in Gyr
    age = 0.5      # age of Universe in units of 1/H0
    age_Gyr = 0.0  # value of age in Gyr
    zage = 0.1     # age of Universe at redshift z in units of 1/H0
    zage_Gyr = 0.0 # value of zage in Gyr
    DCMR = 0.0     # comoving radial distance in units of c/H0
    DCMR_Mpc = 0.0 
    DCMR_Gyr = 0.0
    DA = 0.0       # angular size distance
    DA_Mpc = 0.0
    DA_Gyr = 0.0
    kpc_DA = 0.0
    DL = 0.0       # luminosity distance
    DL_Mpc = 0.0
    DL_Gyr = 0.0   # DL in units of billions of light years
    V_Gpc = 0.0
    a = 1.0        # 1/(1+z), the scale factor of the Universe
    az = 0.5       # 1/(1+z(object))

    h = H0/100.
    WR = 4.165E-5/(h*h)   # includes 3 massless neutrino species, T0 = 2.72528
    WK = 1-WM-WR-WV
    az = np.array(1.0/(1+1.0*z))
    age = 0.
    n=1000     
    # number of points in integrals
    for i in range(n):
        a = az*(i+0.5)/n
        adot = np.sqrt(WK+(WM/a)+(WR/(a*a))+(WV*a*a))
        age = age + 1./adot

    zage = az*age/n
    zage_Gyr = (Tyr/H0)*zage
    DTT = 0.0
    DCMR = 0.0

    # do integral over a=1/(1+z) from az to 1 in n steps, midpoint rule
    for i in range(n):
        a = az+(1-az)*(i+0.5)/n
        adot = np.sqrt(WK+(WM/a)+(WR/(a*a))+(WV*a*a))
        DTT = DTT + 1./adot
        DCMR = DCMR + 1./(a*adot)

    DTT = (1.-az)*DTT/n
    DCMR = (1.-az)*DCMR/n
    age = DTT+zage
    age_Gyr = age*(Tyr/H0)
    DTT_Gyr = (Tyr/H0)*DTT
    DCMR_Gyr = (Tyr/H0)*DCMR
    DCMR_Mpc = (c/H0)*DCMR

    # tangential comoving distance

    ratio = 1.00
    x = np.sqrt(abs(WK))*DCMR
    if x.all() > 0.1:
        if WK > 0:
            ratio =  0.5*(np.exp(x)-np.exp(-x))/x 
        else:
            ratio = np.sin(x)/x
    else:
        y = x*x
        if WK < 0: 
            y = -y
        ratio = 1. + y/6. + y*y/120.

    DCMT = ratio*DCMR
    DA = az*DCMT
    DA_Mpc = (c/H0)*DA
    kpc_DA = DA_Mpc/206.264806
    DA_Gyr = (Tyr/H0)*DA
    DL = DA/(az*az)
    DL_Mpc = (c/H0)*DL
    DL_Gyr = (Tyr/H0)*DL
    
    return DA_Mpc


def great_circle_distance(ra1, dec1, ra2, dec2):
    dLat = (dec1-dec2)
    dLng = (ra2-ra1)

    val = np.sin(dLat/2) * np.sin(dLat/2) + np.sin(dLng/2) * np.sin(dLng/2) * np.cos(dec1) * np.cos(dec2)    
    ang = 2 * np.arctan2(np.sqrt(val), np.sqrt(1-val))
    return ang





  
from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

__version__ = "0.0.3"
setup(
    name="galaxy_environment",
    version=__version__,
    description="Measuring Galaxies Environments",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Kayleigh Meneghini",
    author_email="kayleighmeneghini@gmail.com",
    url="https://github.com/kaykeigh/galaxy-environment",
    packages=["galaxy_environment"],
    install_requires=[
        "numpy",
        "pandas"
    ],
    license="Apache License 2.0",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.5",
)


from . import distances

import math
import warnings
warnings.filterwarnings("ignore")

import numpy as np
import pandas as pd


def createDataFrame(df, estimator = ['knn', 'fixed'], kn=[1, 7, 10, 50], radius=[0.25,1,2,5,8,10]):
    '''
    Creates a pandas dataframe for the environmental estimators.
    
    Parameters:
    -----------
    
    df: Pandas dataframe containing the sample of galaxies;
    
    estimator: str
    'knn' for the nearest neighbors method
    'fixed' for fixed apertures method.
    
    kn: list
    List containing k values for calculating distances to k-neighbors.
    
    radius: list
    List containing the projected radius values in Mpc.
    
    '''
    
    if estimator == 'knn': 
        df_environment = pd.DataFrame()
        df_environment = (df_environment
                   .assign(ID = np.repeat(0,len(df)), 
                          ra = np.repeat(0,len(df)),
                          dec = np.repeat(0,len(df)),
                          z = np.repeat(0,len(df))))
    
        for k in range(0, len(kn)):
             df_environment["sigma_k" + str(kn[k])] = np.repeat(0,len(df))

        df_environment = (df_environment
                    .assign(n_neighborhood = np.repeat(0,len(df))))

    elif estimator == 'fixed':
        df_environment = pd.DataFrame()
        df_environment = (df_environment
                .assign(ID = np.repeat(0,len(df)), 
                        ra = np.repeat(0,len(df)),
                        dec = np.repeat(0,len(df)),
                        z = np.repeat(0,len(df))))
        
        for k in range(0, len(radius)):
            df_environment["sigma_r" + str(radius[k])] = np.repeat(0,len(df))
                
        df_environment = (df_environment
                .assign(n_neighborhood = np.repeat(0,len(df))))

    
    return df_environment

def knn(df, id_column = 'ID', ra_column='RA', dec_column='DEC' , degrees = True, z_column ='z', delta_z=None, kn=[1, 7, 10, 50]):
    ''' 
    For each galaxy in the sample, we choose values for k, 
    which is the number of neighbors around a range of z. 
    The distance to each k neighbor is calculated and 
    the surface density of galaxies in that radius is estimated. 
   
   
    Parameters:
    -----------
    df = Pandas dataframe
    Each line represents a galaxy with its respective coordinates (RA and Dec) and redshift (z);
    
    ra_column = str
    Column name in the dataframe containing the right ascension coordinate in degrees;
    
    dec_column = str
    Column name in the dataframe containing the declination coordinate in degrees;
    
    degrees = Boolean
    True by default. 
    If your coordinates are in radians, pass the value False for the parameter;
    
    z_column = str
    Column name in the dataframe containing the redshift;
    
    delta_z = float
    None by default, it will be taken two standard deviations of z column.  
    If you prefer, you can pass a float for this parameter;
    
    kn = list
    List containing k values for calculating distances to k-neighbors.
    
    
     
    ''' 
 
    if degrees == True: 
        ra = df[ra_column].values * (math.pi/180)
        dec = df[dec_column].values * (math.pi/180)
    else: 
        ra = df[ra_column].values
        dec = df[dec_column].values 
        
    ids = df[id_column].values
    zs = df[z_column].values
    
    if delta_z is None:
        delta_z = np.std(zs).round(3)
    
    df_environment = createDataFrame(ra, estimator = 'knn', kn = kn )
    distd = distances.cosmological_distances(zs)
    dens_knn=np.tile(0.0,len(kn))
  

    if len(ra) > 10000:
        print('Hmm, I just checked here and it seems your sample is larger than 10k rows... \n')
        print('Go grab some coffee, it may take a few minutes!')
    else:
        print('It will be ready in less than a minute... You are lucky!')

    for i in range(0, len(ra)):
    #definindo intervalo de redshift
        z = zs[i]
        z1 = z-2*delta_z
        z2 = z+2*delta_z
        ra1 = ra[i]
        dec1 = dec[i]
        dist = distd[i]

        # pegando o subgrupo de galáxias nesse intervalo
        ra2 = ra[np.logical_and(zs > z1, zs <= z2)]
        dec2 = dec[np.logical_and(zs > z1, zs <= z2)]

        
        df_environment['ID'].iloc[i] = ids[i]
        df_environment['n_neighborhood'].iloc[i] = len(ra2)
        df_environment['ra'].iloc[i] = ra1
        df_environment['dec'].iloc[i] = dec1
        df_environment['z'].iloc[i] = z

        d = distances.great_circle_distance(ra1, dec1, ra2, dec2)*dist
        
        # distance to k-neighbours
        indice = np.argsort(d)

        #density
        for k in range(0, len(kn)):
            try: 
                dens_knn[k] = kn[k]/(math.pi*d[indice[kn[k]]]**2)

            except:
                dens_knn[k]=0

            df_environment["sigma_k" + str(kn[k])].iloc[i] = dens_knn[k]
 

    return df_environment

def fixedApertures(df, id_column = 'ID',  ra_column='RA', dec_column='DEC' , degrees = True, z_column ='z', delta_z=None, 
                   radius =[0.25,1,2,5,8,10]):
    ''' 
    This method defines a scale for the environment, 
    We calculate the density by counting the number of galaxies 
    within a fixed radius in Mpc around a range of z.
   
   
    Parameters:
    -----------
    df = Pandas dataframe
    Each line represents a galaxy with its respective coordinates (RA and Dec) and redshift (z);
    
    ra_column = str
    Column name in the dataframe containing the right ascension coordinate in degrees;
    
    dec_column = str
    Column name in the dataframe containing the declination coordinate in degrees;
    
    degrees = Boolean
    True by default. 
    If your coordinates are in radians, pass the value False for the parameter;
    
    z_column = str
    Column name in the dataframe containing the redshift;
    
    delta_z = float
    None by default, it will be taken two standard deviations of z column. 
    If you prefer, you can pass a float for this parameter;
    
    radius = list
    List containing the projected radius values in Mpc.
    
    '''
    
    if degrees: 
        ra = df[ra_column].values * (math.pi/180)
        dec = df[dec_column].values * (math.pi/180)
    else: 
        ra = df[ra_column].values
        dec = df[dec_column].values 
        
    ids = df[id_column].values
    zs = df[z_column].values
    
    if delta_z is None:
        delta_z = np.std(zs).round(3)
    
    df_environment = createDataFrame(ra, estimator = 'fixed', radius = radius)
    distd = distances.cosmological_distances(zs)
    dens_radius=np.tile(0.0,len(radius))

    if len(ra) > 10000:
        print('Hmm, I just checked here and it seems your sample is larger than 10k rows... \n')
        print('Go grab some coffee! It may take a while...')
    else:
        print('It will be ready in less than a minute... You are lucky!')

    for i in range(0, len(ra)):
    #definindo intervalo de redshift
        z = zs[i]
        z1 = z-2*delta_z
        z2 = z+2*delta_z
        ra1 = ra[i]
        dec1 = dec[i]
        dist = distd[i]

        # pegando o subgrupo de galáxias nesse intervalo
        ra2 = ra[np.logical_and(zs > z1, zs <= z2)]
        dec2 = dec[np.logical_and(zs > z1, zs <= z2)]

        df_environment['ID'].iloc[i] = ids[i]
        df_environment['n_neighborhood'].iloc[i] = len(ra2)
        df_environment['ra'].iloc[i] = ra1
        df_environment['dec'].iloc[i] = dec1
        df_environment['z'].iloc[i] = z

        d = distances.great_circle_distance(ra1, dec1, ra2, dec2)*dist
        

        #density 
        for k in range(0, len(radius)):
            dens_radius[k] = len(d[d < radius[k]])/(math.pi*radius[k]**2)

            df_environment["sigma_r" + str(radius[k])].iloc[i] = dens_radius[k]

    return df_environment